import { __extends } from 'tslib';
import { FormGroup, FormControl, FormBuilder } from '@angular/forms';
import { BehaviorSubject } from 'rxjs/BehaviorSubject';
import { MetadataStorage, Validator, getFromContainer, ValidationTypes, validateSync } from 'class-validator';
import { classToClass, plainToClass } from 'class-transformer';
import 'reflect-metadata';
import { mergeWith, cloneDeep } from 'lodash-es';

var DynamicFormGroup = /** @class */ (function (_super) {
    __extends(DynamicFormGroup, _super);
    function DynamicFormGroup(factoryModel, fields) {
        var _this = _super.call(this, {}) || this;
        _this.factoryModel = factoryModel;
        _this.fields = fields;
        _this.customValidateErrors = new BehaviorSubject({});
        _this.fields = _this.onlyFields(_this.fields);
        return _this;
    }
    Object.defineProperty(DynamicFormGroup.prototype, "externalErrors", {
        get: function () {
            return this._externalErrors;
        },
        set: function (externalErrors) {
            this._externalErrors = externalErrors;
            this.validate();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DynamicFormGroup.prototype, "validatorOptions", {
        get: function () {
            return this._validatorOptions;
        },
        set: function (validatorOptions) {
            this._validatorOptions = validatorOptions;
            this.validate();
        },
        enumerable: true,
        configurable: true
    });
    DynamicFormGroup.getClassValidators = function (factoryModel, fields, validatorOptions) {
        var validationMetadatas = getFromContainer(MetadataStorage).getTargetValidationMetadatas(factoryModel, '', validatorOptions && validatorOptions.groups ? validatorOptions.groups : undefined);
        var formGroupFields = {};
        var validator = new Validator();
        Object.keys(fields).filter(function (key) { return key.indexOf('__') !== 0; }).forEach(function (key) {
            var formGroupField = formGroupFields[key];
            if (formGroupField === undefined) {
                formGroupField = Array.isArray(fields[key]) ? fields[key] : [];
            }
            if (!Array.isArray(fields[key])) {
                if (fields[key] instanceof DynamicFormGroup) {
                    fields[key].object = fields[key].fields;
                    formGroupField.push(fields[key]);
                }
                else {
                    formGroupField.push(fields[key]);
                }
            }
            validationMetadatas.forEach(function (validationMetadata) {
                if (validationMetadata.propertyName === key) {
                    switch (validationMetadata.type) {
                        default: {
                            for (var typeKey in ValidationTypes) {
                                if (ValidationTypes.hasOwnProperty(typeKey)) {
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] === undefined &&
                                        validationMetadata.type === 'nestedValidation' &&
                                        typeKey === 'NESTED_VALIDATION') {
                                        var nestedValidate = function (c) {
                                            var isValid = (c.parent && c.parent.value) ?
                                                validateSync(c.value, validatorOptions).length === 0 : true;
                                            return isValid ? null : {
                                                nestedValidate: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(nestedValidate);
                                    }
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] === undefined &&
                                        validationMetadata.type === 'customValidation' &&
                                        typeKey === 'CUSTOM_VALIDATION') {
                                        var customValidation = function (c) {
                                            var object = (c.parent instanceof DynamicFormGroup) ?
                                                ((c.parent)).object :
                                                (c.parent ? c.parent.value : {});
                                            if (object) {
                                                object[key] = c.value;
                                            }
                                            var validateErrors = (c.parent && c.parent.value) ?
                                                validateSync(object, validatorOptions) : [];
                                            var isValid = validateErrors.filter(function (error) {
                                                if (error.children.length &&
                                                    error.children.filter(function (children) { return children.property === key; })) {
                                                    return true;
                                                }
                                                return true;
                                            }).length === 0;
                                            return isValid ? null : {
                                                customValidation: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(customValidation);
                                    }
                                    if (validationMetadata.type === ValidationTypes[typeKey] &&
                                        validator[validationMetadata.type] !== undefined) {
                                        var customValidate = function (c) {
                                            if (!c) {
                                                return null;
                                            }
                                            var isValid = c.parent && c.parent.value ? validator[validationMetadata.type](c.value) : true;
                                            return isValid ? null : {
                                                customValidate: {
                                                    valid: false,
                                                    type: validationMetadata.type
                                                }
                                            };
                                        };
                                        formGroupField.push(customValidate);
                                    }
                                }
                            }
                            break;
                        }
                    }
                }
            });
            if (formGroupField[0] instanceof DynamicFormGroup) {
                formGroupFields[key] = formGroupField[0];
            }
            else {
                formGroupFields[key] = [
                    formGroupField[0],
                    formGroupField.filter(function (item, index) { return index !== 0; })
                ];
            }
        });
        return formGroupFields;
    };
    DynamicFormGroup.prototype.onlyFields = function (fields) {
        var _this = this;
        var newFields = {};
        if (fields !== undefined) {
            Object.keys(fields).forEach(function (key) {
                if (fields[key] instanceof DynamicFormGroup) {
                    newFields[key] = _this.onlyFields(((fields[key])).fields);
                }
                else {
                    newFields[key] = fields[key];
                }
            });
        }
        return newFields;
    };
    Object.defineProperty(DynamicFormGroup.prototype, "object", {
        get: function () {
            var _this = this;
            var object = this._object ? this.classToClass(this._object) : new this.factoryModel();
            if (object !== undefined) {
                Object.keys(this.controls).forEach(function (key) {
                    if (_this.controls[key] instanceof DynamicFormGroup) {
                        object[key] = ((_this.controls[key])).object;
                    }
                    else {
                        object[key] = _this.controls[key].value;
                    }
                });
            }
            return this.plainToClass(this.factoryModel, object);
        },
        set: function (object) {
            var _this = this;
            if (object instanceof this.factoryModel) {
                this._object = this.classToClass(object);
            }
            else {
                this._object = this.plainToClass(this.factoryModel, (object));
            }
            Object.keys(this.controls).forEach(function (key) {
                if (_this.controls[key] instanceof DynamicFormGroup) {
                    ((_this.controls[key])).object = _this._object ? _this._object[key] : {};
                }
                else {
                    _this.controls[key].setValue((_this._object && _this._object[key]) ? _this._object[key] : undefined);
                }
            });
        },
        enumerable: true,
        configurable: true
    });
    DynamicFormGroup.prototype.transformValidationErrors = function (errors) {
        var _this = this;
        var customErrors = {};
        errors.forEach(function (error) {
            if (error && error.constraints !== undefined) {
                Object.keys(error.constraints).forEach(function (key) {
                    if (!customErrors[error.property]) {
                        customErrors[error.property] = [];
                    }
                    if (((customErrors[error.property])).indexOf(error.constraints[key]) === -1) {
                        ((customErrors[error.property])).push(error.constraints[key]);
                    }
                });
            }
            if (error.children !== undefined && error.children.length) {
                customErrors[error.property] = _this.transformValidationErrors(error.children);
            }
        });
        return customErrors;
    };
    DynamicFormGroup.prototype.classToClass = function (object) {
        return classToClass(object, { ignoreDecorators: true });
    };
    DynamicFormGroup.prototype.plainToClass = function (cls, plain) {
        return plainToClass(cls, plain, { ignoreDecorators: true });
    };
    DynamicFormGroup.prototype.mergeErrors = function (errors, externalErrors) {
        return mergeWith(errors, externalErrors, function (objValue, srcValue) {
            if (Array.isArray(objValue) &&
                Array.isArray(srcValue) &&
                objValue.filter(function (objItem) { return srcValue.indexOf(objItem) !== -1; }).length === 0) {
                return objValue.concat(srcValue);
            }
        });
    };
    DynamicFormGroup.prototype.validate = function (externalErrors, validatorOptions) {
        if (externalErrors === undefined) {
            externalErrors = cloneDeep(this.externalErrors);
        }
        if (validatorOptions === undefined) {
            validatorOptions = cloneDeep(this.validatorOptions);
        }
        if (!externalErrors) {
            externalErrors = {};
        }
        var errors = validateSync(this.object, validatorOptions);
        var transformedErrors = this.transformValidationErrors(errors);
        var allErrors = this.mergeErrors(externalErrors, transformedErrors);
        this.markAsInvalidForExternalErrors(externalErrors, this.controls);
        this.customValidateErrors.next(allErrors);
    };
    DynamicFormGroup.prototype.markAsInvalidForExternalErrors = function (errors, controls) {
        Object.keys(controls).forEach(function (field) {
            var control = controls[field];
            if (control instanceof FormControl) {
                if (errors && errors[field]) {
                    control.setErrors({ 'externalError': true });
                }
                else {
                    if (control.errors && control.errors["externalError"] === true) {
                        control.setErrors(null);
                    }
                }
            }
            else if (control instanceof DynamicFormGroup) {
                control.markAsInvalidForExternalErrors(errors && errors[field] ? (errors[field]) : {}, control.controls);
            }
        });
    };
    DynamicFormGroup.prototype.validateAllFormFields = function () {
        var _this = this;
        Object.keys(this.controls).forEach(function (field) {
            var control = _this.get(field);
            if (control instanceof FormControl) {
                control.markAsTouched({ onlySelf: true });
            }
            else if (control instanceof DynamicFormGroup) {
                control.validateAllFormFields();
            }
        });
    };
    return DynamicFormGroup;
}(FormGroup));
var DynamicFormBuilder = /** @class */ (function (_super) {
    __extends(DynamicFormBuilder, _super);
    function DynamicFormBuilder() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    DynamicFormBuilder.prototype.group = function (factoryModel, controlsConfig, extra) {
        var _this = this;
        if (extra !== undefined && extra.customValidatorOptions === undefined) {
            extra.customValidatorOptions = { validationError: { target: false } };
        }
        var newControlsConfig;
        if (controlsConfig !== undefined) {
            newControlsConfig = controlsConfig;
        }
        if (controlsConfig === undefined) {
            newControlsConfig = new factoryModel();
            Object.keys(newControlsConfig).forEach(function (key) {
                if (newControlsConfig[key] &&
                    newControlsConfig[key].constructor &&
                    typeof newControlsConfig[key] === 'object' &&
                    (newControlsConfig[key].length === undefined ||
                        (newControlsConfig[key].length !== undefined &&
                            Object.keys(newControlsConfig[key].length).length === newControlsConfig[key].length))) {
                    newControlsConfig[key] = _this.group(newControlsConfig[key].constructor, undefined, extra);
                }
            });
        }
        var dynamicFormGroup = new DynamicFormGroup(factoryModel, newControlsConfig);
        var formGroup = _super.prototype.group.call(this, DynamicFormGroup.getClassValidators(factoryModel, newControlsConfig, (extra && extra.customValidatorOptions) ? extra.customValidatorOptions : undefined), extra);
        Object.keys(formGroup.controls).forEach(function (key) {
            dynamicFormGroup.addControl(key, formGroup.controls[key]);
        });
        dynamicFormGroup.valueChanges.subscribe(function (data) {
            dynamicFormGroup.validate(undefined, (extra && extra.customValidatorOptions) ? extra.customValidatorOptions : undefined);
        });
        return dynamicFormGroup;
    };
    return DynamicFormBuilder;
}(FormBuilder));
var NgxDynamicFormBuilderShareds = [DynamicFormBuilder, DynamicFormGroup];

export { DynamicFormBuilder, DynamicFormGroup, NgxDynamicFormBuilderShareds };
//# sourceMappingURL=ngx-dynamic-form-builder.js.map
